from __future__ import absolute_import
from __future__ import print_function
from unittest.mock import patch
from noaa_sdk import noaa
import pytest

try:
    from unittest.mock import MagicMock
except:
    from mock import MagicMock


def test_instantiation():
    """Test instantiation of NOAA class.
    """
    n = noaa.NOAA()
    assert n._user_agent == n.DEFAULT_USER_AGENT


def test_set_accept():
    """Test setting valid accept string.
    """
    with pytest.raises(Exception) as err:
        noaa.NOAA(accept='test')
    assert str(err.value) == (
        "Invalid format. Available formats are: "
        "["
        "'application/atom+xml', "
        "'application/cap+xml', "
        "'application/geo+json', "
        "'application/json', "
        "'application/ld+json', "
        "'application/vnd.noaa.dwml+xml', "
        "'application/vnd.noaa.obs+xml'"
        "]")


def test_test_user_agent():
    """Test able to set/get user agent.
    """
    n = noaa.NOAA(user_agent='test_agent')
    assert n._user_agent == 'test_agent'
    n.user_agent = 'test_agent2'
    assert n.user_agent == 'test_agent2'


def test_get_request_header():
    """Test request header is set correctly.
    """

    n = noaa.NOAA(user_agent='test_agent', accept=noaa.ACCEPT.CAP)
    request_header = n.get_request_header()
    assert request_header == {
        'User-Agent': 'test_agent',
        'accept': 'application/cap+xml'
    }


@patch('noaa_sdk.util.requests')
def test_make_get_request(mock_requests):
    mock_response_obj = MagicMock()
    mock_response_obj.text = 'mock text'
    mock_response_obj.json = lambda: {"test":"test"}
    mock_response_obj.status_code = 200
    mock_requests.get.return_value = mock_response_obj

    n = noaa.NOAA(user_agent='test_agent')
    res = n.make_get_request(
        'http://test', end_point='test.paulo.com')
    assert res == {"test": "test"}


@patch('noaa_sdk.util.requests')
def test_make_get_request_failed(mock_requests):
    mock_response_obj = MagicMock()
    mock_response_obj.text = 'mock text'
    mock_response_obj.status_code = 500
    mock_response_obj.json = lambda: {"test":"test"}
    mock_requests.get.return_value = mock_response_obj

    with pytest.raises(Exception) as err:
        n = noaa.NOAA(user_agent='test_agent')
        n.make_get_request('http://test')
        assert err == 'Error: end_point is None.'


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_points(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.points('23.44,34.55')
    mock_make_get_request.assert_called_with(
        '/points/23.44,34.55', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_points_with_stations(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.points('23.44,34.55', stations=True)
    mock_make_get_request.assert_called_with(
        '/points/23.44,34.55/stations', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_points_forecast(mock_make_get_request):
    mock_make_get_request.return_value = {
        'properties': {
            'forecast': 'forecast_uri',
            'forecastHourly': 'forecast_hourly_uri'
        }
    }
    n = noaa.NOAA(user_agent='test_agent')
    n.points_forecast(23.44, 34.55, hourly=False)
    mock_make_get_request.assert_any_call(
        uri='forecast_uri', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_points_forecast_with_hourly(mock_make_get_request):
    mock_make_get_request.return_value = {
        'properties': {
            'forecast': 'forecast_hourly_uri',
            'forecastHourly': 'forecast_hourly_uri'
        }
    }
    n = noaa.NOAA(user_agent='test_agent')
    n.points_forecast(23.44, 34.55, hourly=True)
    mock_make_get_request.assert_any_call(
        uri='forecast_hourly_uri', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_stations(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.stations()
    mock_make_get_request.assert_called_with(
        '/stations', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_stations_with_station_id(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.stations(station_id='PAULOSTATION')
    mock_make_get_request.assert_called_with(
        '/stations?id=PAULOSTATION', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_stations_observations(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.stations_observations('PAULOSTATION')
    mock_make_get_request.assert_called_with(
        '/stations/PAULOSTATION/observations', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_stations_observations_with_current(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.stations_observations('PAULOSTATION', current=True)
    mock_make_get_request.assert_called_with(
        '/stations/PAULOSTATION/observations/current',
        end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_stations_observations_with_recordId(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.stations_observations(
        'PAULOSTATION', recordId='2017-01-04T18:54:00+00:00')
    mock_make_get_request.assert_called_with(
        '/stations/PAULOSTATION/observations/2017-01-04T18:54:00+00:00',
        end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_products(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.products('test_id')
    mock_make_get_request.assert_called_with(
        '/products/test_id', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_products_types_with_nothing(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.products_types()
    mock_make_get_request.assert_called_with(
        '/products/types', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_products_types_with_id(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.products_types(type_id='test_id')
    mock_make_get_request.assert_called_with(
        '/products/types/test_id', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_products_types_with_locations_failed(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')

    with pytest.raises(Exception) as err:
        n.products_types(locations=True, location_id='test_location_id')
    assert str(err.value) == ('Error: Missing type id (type_id=None)')


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_products_types_with_locations_and_id(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.products_types(
        locations=True,
        location_id='test_location_id', type_id='test_id')
    mock_make_get_request.assert_called_with(
        '/products/types/test_id/locations/test_location_id',
        end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_products_locations(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.products_locations(
        location_id='test_location_id')
    mock_make_get_request.assert_any_call(
        '/products/locations/test_location_id/types',
        end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_products_locations_without_id(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.products_locations()
    mock_make_get_request.assert_called_with(
        '/products/locations', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_offices(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.offices('test_office_id')
    mock_make_get_request.assert_called_with(
        '/offices/test_office_id', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_zones(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.zones('test_type', 'test_zone_id')
    mock_make_get_request.assert_called_with(
        '/zones/test_type/test_zone_id', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_zones_with_forecast(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.zones('test_type', 'test_zone_id', forecast=True)
    mock_make_get_request.assert_called_with(
        '/zones/test_type/test_zone_id/forecast',
        end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_alerts(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.alerts(
        active=2)
    mock_make_get_request.assert_called_with(
        '/alerts?active=2', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_alerts_with_alert_id(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.alerts(alert_id='test_alert_id')
    mock_make_get_request.assert_called_with(
        '/alerts/test_alert_id', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_active_alerts(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.active_alerts()
    mock_make_get_request.assert_called_with(
        '/alerts/active', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_active_alerts_with_count(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.active_alerts(count=True)
    mock_make_get_request.assert_called_with(
        '/alerts/count', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_active_alerts_with_zone_id(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.active_alerts(zone_id='test_zone_id')
    mock_make_get_request.assert_called_with(
        '/alerts/active/zone/test_zone_id', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_active_alerts_with_area(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.active_alerts(area='nyc')
    mock_make_get_request.assert_called_with(
        '/alerts/active/area/nyc', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.NOAA.make_get_request')
def test_active_alerts_with_region(mock_make_get_request):
    mock_make_get_request.return_value = None
    n = noaa.NOAA(user_agent='test_agent')
    n.active_alerts(region='test_region')
    mock_make_get_request.assert_called_with(
        '/alerts/active/region/test_region', end_point=n.DEFAULT_END_POINT)


@patch('noaa_sdk.noaa.OSM.make_get_request')
def test_osm_get_lat_lon_by_postalcode_country(mock_make_get_request):
    mock_make_get_request.return_value = [{'lat': 56.34, 'lon': 12.78}]
    n = noaa.OSM()
    lat, lon = n.get_lat_lon_by_postalcode_country('11365', 'US')
    mock_make_get_request.assert_called_with(
        '/search?postalcode=11365&country=US&format=json',
        end_point=n.OSM_ENDPOINT)
    assert lat == 56.34
    assert lon == 12.78


@patch('noaa_sdk.noaa.OSM.make_get_request')
def test_osm_get_lat_lon_by_postalcode_country_failed(mock_make_get_request):
    mock_make_get_request.return_value = []
    with pytest.raises(Exception) as err:
        n = noaa.OSM()
        n.get_lat_lon_by_postalcode_country('11365', 'US')
        mock_make_get_request.assert_called_with(
            '/search?postalcode=11365&country=US&format=json',
            end_point=n.OSM_ENDPOINT)
        assert err == 'No response from: {}'.format(n.OSM_ENDPOINT)


@patch('noaa_sdk.noaa.OSM.make_get_request')
def test_osm_get_postalcode_country_by_lan_lon(mock_make_get_request):
    mock_make_get_request.return_value = {
        'address': {
            'postcode': '11375', 'country_code': 'US'}
    }
    n = noaa.OSM()
    postcode, country = n.get_postalcode_country_by_lan_lon(23.22, 33.33)
    mock_make_get_request.assert_called_with(
        '/reverse?lat=23.22&lon=33.33&addressdetails=1&format=json',
        end_point=n.OSM_ENDPOINT)
    assert postcode == '11375'
    assert country == 'US'


@patch('noaa_sdk.noaa.OSM.make_get_request')
def test_osm_get_postalcode_country_by_lan_lon_failed(mock_make_get_request):
    mock_make_get_request.return_value = {}
    with pytest.raises(Exception) as err:
        n = noaa.OSM()
        n.get_postalcode_country_by_lan_lon(23.22, 33.33)
        mock_make_get_request.assert_called_with(
            '/reverse?lat=23.22&lon=33.33&addressdetails=1&format=json',
            end_point=n.OSM_ENDPOINT)
        assert err == 'No response from: {}'.format(n.OSM_ENDPOINT)



from urllib.parse import urlencode
import requests

from util import UTIL


class NCDC(UTIL):
    """Main class for getting data from NCDC.
    Documentation: https://www.ncdc.noaa.gov/cdo-web/webservices/v2
    """

    DEFAULT_END_POINT = 'www.ncdc.noaa.gov'
    DEFAULT_USER_AGENT = 'Test (your@email.com)'

    END_POINTS = {
        'datasets', 'datacategories',
        'datatypes', 'locationcategories',
        'locations', 'stations', 'data'}

    def __init__(self, token, user_agent=None, accept=None, show_uri=False):
        """Constructor.

        Args:
            token (str): token for api.
            user_agent (str[optional]): user agent specified in the header.
            accept (str[optional]): accept string specified in the header.
            show_uri (boolean[optional]): True for showing the
                actual url with query string being sent for requesting data.
        """
        if not token:
            raise Exception('Error: missing token.')
        if not user_agent:
            user_agent = self.DEFAULT_USER_AGENT

        self._token = token
        super().__init__(
            user_agent=user_agent, accept=accept,
            show_uri=show_uri)

    def make_get_request(self, uri, header=None, end_point=None):
        """Encapsulate code for GET request.

        Args:
            uri (str): full get url with query string.
            header (dict): request header.
            end_point (str): end point host.

        Returns:
            dict: dictionary response.
        """

        if self._show_uri:
            print('Calling: {}'.format(uri))
        if not end_point:
            raise Exception('Error: end_point is None.')

        res = requests.get(
            'https://{}/{}'.format(self.DEFAULT_END_POINT, uri),
            headers=header)
        if res.status_code == 200:
            return res.json()
        raise Exception('Error: {} {}'.format(res.status_code, res.reason))

    def get_request_header(self):
        """Get required headers.

        Args:
            format (str): content type.

        Returns:
            dec: headers with access token.
        """

        return {
            'token': self._token
        }

    def request_end_point(self, end_point, **params):

        return self.make_get_request(
            end_point=self.DEFAULT_END_POINT,
            header=self.get_request_header(),
            uri='cdo-web/api/v2/{}?{}'.format(end_point, urlencode(params)))

    def datasets(self, **params):
        """ Request datasets endpoint.

        Args:
            datatypeid (str[optional]):	eg. ACMM. Accepts a valid data type id
                or a chain of data type ids seperated by ampersands.
                Datasets returned will contain all of the data type(s) specified.
            locationid (str[optional]):	eg. FIPS:37. Accepts a valid location id
                or a chain of location ids seperated by ampersands. Datasets
                returned will contain data for the location(s) specified.
            stationid (str[optional]): eg. COOP:010957. Accepts a valid station id
                or a chain of of station ids seperated by ampersands.
                Datasets returned will contain data for the station(s) specified.
            startdate (str[optional]): eg. 1970-10-03. Accepts valid ISO formated
                date (yyyy-mm-dd). Datasets returned will have data after the
                specified date. Paramater can be use independently of enddate
            enddate	(str[optional]): eg. 2012-09-10. Accepts valid ISO formated date (yyyy-mm-dd).
                Datasets returned will have data before the specified date.
                Paramater can be use independently of startdate
            sortfield (str[optional]): eg.name. The field to sort results by.
                Supports id, name, mindate, maxdate, and datacoverage fields.
            sortorder (str[optional]): eg. desc. Which order to sort by, asc or desc.
                Defaults to asc
            limit (int[optional]): eg.42. Defaults to 25, limits the number of
                results in the response. Maximum is 1000
            offset (str[optional]): eg.24. Defaults to 0, used to offset the resultlist.
                The example would begin with record 24.
        """

        return self.request_end_point('datasets', **params)



class ACCEPT(object):
    """Encapsulate all accept strings."""

    GEOJSON = 'application/geo+json'
    JSONLD = 'application/ld+json'
    DWML = 'application/vnd.noaa.dwml+xml'
    OXML = 'application/vnd.noaa.obs+xml'
    CAP = 'application/cap+xml'
    ATOM = 'application/atom+xml'
    JSON = 'application/json'



import requests
import time
from datetime import datetime
from functools import wraps

from noaa_sdk.accept import ACCEPT


class UTIL(object):
    """Utility class for making requests."""

    def __init__(self, user_agent='', accept=None, show_uri=False):
        """Constructor.

        Args:
            user_agent (str[optional]): user agent specified in the header.
            accept (str[optional]): accept string specified in the header.
        """
        self._show_uri = show_uri
        self._user_agent = user_agent

        if accept:
            accepts = [getattr(ACCEPT, i)
                       for i in dir(ACCEPT) if '__' not in i]
            accepts = sorted(accepts)
            if accept not in accepts:
                raise Exception(
                    'Invalid format. '
                    'Available formats are: {}'.format(accepts))
            self._accept = accept

    def _retry_request_decorator(max_retries):
        def _retry_request_sub_decorator(request):
            @wraps(request)
            def wrapper(*args, **kargs):
                status_code = ''
                retry = 0
                fib_num_a = 1
                fib_num_b = 1
                response = request(*args, **kargs)
                status_code = response.status_code

                while retry <= max_retries and (status_code == '' or status_code == 500):
                    response = request(*args, **kargs)
                    status_code = response.status_code
                    new_interval = fib_num_b + fib_num_a
                    fib_num_a = fib_num_b
                    time.sleep(new_interval)
                    fib_num_b = new_interval
                    retry += 1

                if retry > max_retries:
                    raise Exception(
                        'Maximum retries exceeded. Response object dump: {}'.format(
                            response))
                return response

            return wrapper

        return _retry_request_sub_decorator

    @property
    def show_uri(self):
        return self._show_uri

    @show_uri.setter
    def show_uri(self, value):
        self._show_uri = value

    @property
    def accept(self):
        return self._accept

    @accept.setter
    def accept(self, value):
        self._accept = value

    @property
    def user_agent(self):
        return self._user_agent

    @user_agent.setter
    def user_agent(self, value):
        self._user_agent = value

    def get_request_header(self):
        """Get required headers.

        Args:
            format (str): content type.

        Returns:
            dec: headers with access token.
        """

        return {
            'User-Agent': self._user_agent,
            'accept': self._accept
        }

    @_retry_request_decorator(10)
    def _get(self, end_point, uri, header):
        return requests.get(
            'https://{}/{}'.format(end_point, uri), headers=header)

    def make_get_request(self, uri, header=None, end_point=None):
        """Encapsulate code for GET request.

        Args:
            uri (str): full get url with query string.
            header (dict): request header.
            end_point (str): end point host.

        Returns:
            dict: dictionary response.
        """

        if self._show_uri:
            print('Calling: {}'.format(uri))
        if not header:
            header = self.get_request_header()
        if not end_point:
            raise Exception('Error: end_point is None.')

        if 'http://' in uri or 'https://' in uri:
            uri = uri.replace('http://', '').replace('https://', '')
            end_point = uri.split('/')[0]
            uri = uri.replace(end_point, '')

        res = self._get(end_point, uri, header)

        return res.json()

    def parse_param_timestamp(self, str_date_time):
        """Parse string to datetime object.

        Args:
            str_date_time (str): date time 3 different formats:
                '%Y-%m-%dT%H:%M:%SZ' | '%Y-%m-%d' | '%Y-%m-%d %H:%M:%S'
        Returns:
            datetime object.
        """
        formats = [
            '%Y-%m-%dT%H:%M:%SZ',
            '%Y-%m-%d',
            '%Y-%m-%d %H:%M:%S'
        ]

        for format in formats:
            try:
                return datetime.strptime(str_date_time, format)
            except Exception as err:
                continue

        raise Exception(
            "Error: start and end must have "
            "format '%Y-%m-%dT%H:%M:%SZ' | '%Y-%m-%d' | '%Y-%m-%d %H:%M:%S'")

    def parse_response_timestamp(self, str_date_time):
        """Parse string to datetime object.

        Args:
            str_date_time (str): date time in format (YYYY-MM-DD)
        Returns:
            datetime object.
        """
        return datetime.strptime(str_date_time, '%Y-%m-%dT%H:%M:%S+00:00')



from setuptools import find_packages
from distutils.core import setup

with open('README.rst') as file:
    long_description = file.read()

setup(name='noaa-sdk',
      version='0.1.12a',
      description='NOAA API (V3) Python 3 SDK.',
      install_requires=[
          'httplib2==0.10.3',
          'urllib3==1.22',
          'requests'
      ],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.4'
      ],
      keywords=(
          'NOAA noaa weather public v3 api sdk osm postalcode country postcode'),
      url='https://github.com/paulokuong/noaa',
      author='Paulo Kuong',
      author_email='paulo.kuong@gmail.com',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      long_description=long_description)



"""
API Wrapper for NOAA API V3
===========================
For more detailed information about NOAA API,
visit: https://forecast-v3.weather.gov/documentation

Geoencoding is made possible by Open Street Map (© OpenStreetMap contributors)
For copyright information, visit: https://www.openstreetmap.org/copyright


"""

import json
from urllib.parse import urlencode

from noaa_sdk.util import UTIL
from noaa_sdk.accept import ACCEPT


class OSM(UTIL):
    """
    Make request to Open Street Map nominatim Api.
    ==============================================
    © OpenStreetMap contributors

    Open Street Map, you guys are awesome, don't block me please.
    """

    OSM_ENDPOINT = 'nominatim.openstreetmap.org'

    def __init__(self, show_uri=False):
        """Constructor.
        """
        self._user_agent = 'pypi noaa_sdk'
        self._accept = ACCEPT.JSON
        super().__init__(
            user_agent=self._user_agent, accept=ACCEPT.JSON,
            show_uri=show_uri)

    def get_lat_lon_by_postalcode_country(self, postalcode, country):
        """Get latitude and longitude coordinate from postalcode
        and country code.

        Args:
            postalcode (str): postal code.
            country (str): 2 letter country code.
        Returns:
            tuple: tuple of latitude and longitude.
        """

        res = self.make_get_request(
            '/search?postalcode={}&country={}&format=json'.format(
                postalcode, country), end_point=self.OSM_ENDPOINT)
        return float(res[0]['lat']), float(res[0]['lon'])

    def get_postalcode_country_by_lan_lon(self, lat, lon):
        """Get postalcode and country code by latitude and longitude.

        Args:
            lat (float): latitude.
            lon (float): longitude.
        Returns:
            tuple: tuple of postalcode and country code.
        """
        res = self.make_get_request(
            '/reverse?lat={}&lon={}&addressdetails=1&format=json'.format(
                lat, lon),
            end_point=self.OSM_ENDPOINT)
        if 'address' not in res:
            raise Exception('No address found.')

        if 'country_code' not in res['address']:
            raise Exception('No country code found.')

        if 'postcode' not in res['address']:
            raise Exception('No postal code found.')

        return res['address']['postcode'], res['address']['country_code']


class NOAA(UTIL):
    """Main class for getting data from NOAA."""

    DEFAULT_END_POINT = 'api.weather.gov'
    DEFAULT_USER_AGENT = 'Test (your@email.com)'

    def __init__(self, user_agent=None, accept=None, show_uri=False):
        """Constructor.

        Args:
            user_agent (str[optional]): user agent specified in the header.
            accept (str[optional]): accept string specified in the header.
            show_uri (boolean[optional]): True for showing the
                actual url with query string being sent for requesting data.
        """
        if not user_agent:
            user_agent = self.DEFAULT_USER_AGENT
        if not accept:
            accept = ACCEPT.GEOJSON

        super().__init__(
            user_agent=user_agent, accept=accept,
            show_uri=show_uri)
        self._osm = OSM()

    def get_forecasts(self, postal_code, country, hourly=False):
        """Get forecasts by postal code and country code.

        Args:
            postalcode (str): postal code.
            country (str): 2 letter country code.
            hourly (boolean[optional]): True for getting hourly forecast.
        Returns:
            list: list of weather forecasts.
        """

        lat, lon = self._osm.get_lat_lon_by_postalcode_country(postal_code, country)
        res = self.points_forecast(lat, lon, hourly)
        if 'properties' in res and 'periods' in res['properties']:
            return res['properties']['periods']
        return []

    def get_observations(
            self, postalcode, country, start=None, end=None, num_of_stations=1):
        """Get all nearest station observations by postalcode and
        country code.

           This is an extensive functionality, aligning data
           from Open Street Map to enable postal code and country code
           params for weather observation data.

        Args:
            postalcode (str): postal code.
            country (str): 2 letter country code.
            start (str[optional]): start date of observation
                (eg. '%Y-%m-%dT%H:%M:%SZ' | '%Y-%m-%d' | '%Y-%m-%d %H:%M:%S').
            end (str[optional]): end date of observation
                (eg. '%Y-%m-%dT%H:%M:%SZ' | '%Y-%m-%d' | '%Y-%m-%d %H:%M:%S').
            num_of_stations (int[optional]): get observations from the
                nearest x stations. (Put -1 of wants to get all stations.)
        Returns:
            generator: generator of dictionaries of observations with the
            following variables: 'relativeHumidity', 'presentWeather',
            'minTemperatureLast24Hours', 'windGust', 'precipitationLast6Hours',
            'cloudLayers', 'dewpoint', 'temperature', 'windDirection',
            'rawMessage', '@type', 'precipitationLast3Hours',
            'visibility', 'icon', 'barometricPressure', 'windChill',
            'station', 'textDescription', 'seaLevelPressure', '@id',
            'timestamp', 'maxTemperatureLast24Hours', 'precipitationLastHour',
            'heatIndex', 'windSpeed', 'elevation'
        """

        stations_observations_params = {}
        if start:
            stations_observations_params['start'] = start
        if end:
            stations_observations_params['end'] = end

        lat, lon = self._osm.get_lat_lon_by_postalcode_country(postalcode, country)
        points_res = self.points(
            '{},{}'.format(round(lat, 4), round(lon, 4)))

        if 'properties' not in points_res or 'observationStations' not in points_res['properties']:
            raise Exception('Error: No Observation Stations found.')
        stations = self.make_get_request(
            uri=points_res['properties']['observationStations'],
            end_point=self.DEFAULT_END_POINT)['observationStations']

        for index, station in enumerate(stations):
            if num_of_stations > 0 and num_of_stations <= index:
                break
            station_id = station.split('/')[-1]
            response = self.stations_observations(
                station_id=station_id, **stations_observations_params)

            for observation in response['features']:
                yield observation.get('properties')

    def get_observations_by_postalcode_country(
            self, postalcode, country, start=None, end=None, num_of_stations=1):
        """Deprecated. Please use method get_observations."""
        return self.get_observations(postalcode, country, start, end, num_of_stations)

    def points(self, point, stations=False):
        """Metadata about a point.
        This is the primary endpoint for forecast information for a location.
        It contains linked data for the forecast, the hourly forecast,
        observation and other information. It also shows stations nearest to a point
        in order of distance.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            point (str): lat,long.
            stations (boolean): True for finding stations.
        Returns:
            json: json response from api.
        """

        if stations:
            return self.make_get_request(
                "/points/{point}/stations".format(point=point),
                end_point=self.DEFAULT_END_POINT)
        return self.make_get_request(
            "/points/{point}".format(point=point),
            end_point=self.DEFAULT_END_POINT)

    def points_forecast(self, lat, long, hourly=False):
        """Get observation data from a weather station.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            lat (float): latitude of the weather station coordinate.
            long (float): longitude of the weather station coordinate.
            hourly (boolean[optional]): True for getting hourly forecast.
        Returns:
            json: json response from api.
        """

        points = self.make_get_request(
            "/points/{lat},{long}".format(
                lat=lat, long=long), end_point=self.DEFAULT_END_POINT)
        uri = points['properties']['forecast']
        if hourly:
            uri = points['properties']['forecastHourly']

        return self.make_get_request(
            uri=uri, end_point=self.DEFAULT_END_POINT)

    def stations(self, **params):
        """Get list of US weather stations and their metadata.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            station_id (str[optional]): station id.
            state (str[optional]): 2 letters state code.
            limit (int[optional]): limit of results.
        Returns:
            json: json response from api.
        """
        if len(params) > 0:
            if 'station_id' in params:
                params['id'] = params['station_id']
                del params['station_id']
            return self.make_get_request(
                "/stations?{query_string}".format(
                    query_string=urlencode(params)),
                end_point=self.DEFAULT_END_POINT)
        return self.make_get_request(
            "/stations", end_point=self.DEFAULT_END_POINT)

    def stations_observations(self, station_id, **params):
        """Get observation data from specific station.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            station_id (str): station id.
            start (str[optional]): start date of observation
                (eg. '%Y-%m-%dT%H:%M:%SZ' | '%Y-%m-%d' | '%Y-%m-%d %H:%M:%S').
            end (str[optional]): end date of observation
                (eg. '%Y-%m-%dT%H:%M:%SZ' | '%Y-%m-%d' | '%Y-%m-%d %H:%M:%S').
            limit (int[optional]): limit of results.
            current (bool[optional]): True if needs current observations.
            recordId (str[optional]): recordId, Record Id (ISO8601DateTime)
        Returns:
            json: json response from api.
        """
        if len(params) > 0:
            if 'recordId' in params:
                return self.make_get_request(
                    '/stations/{stationId}/observations/{recordId}'.format(
                        stationId=station_id, recordId=params['recordId']),
                    end_point=self.DEFAULT_END_POINT)
            if 'current' in params:
                return self.make_get_request(
                    '/stations/{stationId}/observations/current'.format(
                        stationId=station_id),
                    end_point=self.DEFAULT_END_POINT)

            observations = self.make_get_request(
                "/stations/{stationId}/observations".format(stationId=station_id),
                end_point=self.DEFAULT_END_POINT)

            observations = observations['features']

            # There is an issue on NOAA's side which causes start and end params
            # not enable to filter anything (return empty results). So,
            # for now, there is no chance but to fetch everything and filter
            # the data below.

            if 'start' in params:
                observations = [
                    i for i in observations
                    if self.parse_response_timestamp(
                        i['properties']['timestamp']) >= self.parse_param_timestamp(
                        params['start'])]
            if 'end' in params:
                observations = [
                    i for i in observations
                    if self.parse_response_timestamp(
                        i['properties']['timestamp']) <= self.parse_param_timestamp(
                        params['end'])]
            return observations

        return self.make_get_request(
            "/stations/{stationId}/observations".format(stationId=station_id),
            end_point=self.DEFAULT_END_POINT)

    def products(self, id):
        """Get data of a product.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            id (str): product id.
        Returns:
            json: json response from api.
        """
        return self.make_get_request(
            "/products/{productId}".format(productId=id),
            end_point=self.DEFAULT_END_POINT)

    def products_types(self, **params):
        """Get a list of product types with an active product.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            type_id (str): an id of a valid product type
            locations (boolean[optional]): True to get a list of
                locations that have issues products for a type.
            location_id (str): location id.
        Returns:
            json: json response from api.
        """
        if 'type_id' in params and 'locations' not in params:
            return self.make_get_request(
                "/products/types/{type_id}".format(type_id=params['type_id']),
                end_point=self.DEFAULT_END_POINT)
        elif 'locations' in params:
            if 'type_id' not in params:
                raise Exception('Error: Missing type id (type_id=None)')
            if 'location_id' in params:
                return self.make_get_request(
                    ('/products/types/{type_id}/locations/'
                     '{location_id}').format(
                        type_id=params['type_id'],
                        location_id=params['location_id']),
                    end_point=self.DEFAULT_END_POINT)
            else:
                return self.make_get_request(
                    "/products/types/{type_id}/locations".format(
                        type_id=params['type_id']),
                    end_point=self.DEFAULT_END_POINT)
        return self.make_get_request(
            "/products/types",
            end_point=self.DEFAULT_END_POINT)

    def products_locations(self, **params):
        """A list of locations with active products.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            location_id (str): location id.
        Returns:
            json: json response from api.
        """
        if 'location_id' in params:
            return self.make_get_request(
                "/products/locations/{locationId}/types".format(
                    locationId=params['location_id']),
                end_point=self.DEFAULT_END_POINT)
        return self.make_get_request(
            "/products/locations",
            end_point=self.DEFAULT_END_POINT)

    def offices(self, office_id):
        """Metadata about a Weather Office.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            office_id (str): office id.
        Returns:
            json: json response from api.
        """
        return self.make_get_request("/offices/{office_id}".format(
            office_id=office_id), end_point=self.DEFAULT_END_POINT)

    def zones(self, type, zone_id, forecast=False):
        """Metadata for a zone and forecast data for zone.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            type (str): a valid zone type (list forthcoming).
            zone_id (str): a zone id (list forthcoming).
            forecast (bool): True to show forecast data of the zone.
        Returns:
            json: json response from api.
        """
        if forecast:
            return self.make_get_request(
                "/zones/{type}/{zone_id}/forecast".format(
                    type=type, zone_id=zone_id),
                end_point=self.DEFAULT_END_POINT)
        return self.make_get_request("/zones/{type}/{zone_id}".format(
            type=type, zone_id=zone_id), end_point=self.DEFAULT_END_POINT)

    def alerts(self, **params):
        """A list of alerts that can be filtered by parameters.
        If no parameters are provided, then all alerts are returned.
        The ATOM format returns items in CAP-ATOM.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            alert_id (str): alert id.
            active (int): Active alerts (1 or 0).
            start (str): Start time (ISO8601DateTime).
            end (str): End time (ISO8601DateTime).
            status (str): Event status (alert, update, cancel).
            type (str): Event type (list forthcoming).
            zone_type (str): Zone type (land or marine).
            point (str): Point (latitude,longitude).
            region (str): Region code (list forthcoming).
            state (str): State/marine code (list forthcoming).
            zone (str): Zone Id (forecast or county, list forthcoming).
            urgency (str): Urgency (expected, immediate).
            severity (str): Severity (minor, moderate, severe).
            certainty (str): Certainty (likely, observed).
            limit (int) Limit (an integer).
        Returns:
            json: json response from api.
        """
        if 'alert_id' in params:
            return self.make_get_request(
                "/alerts/{alert_id}".format(alert_id=params['alert_id']),
                end_point=self.DEFAULT_END_POINT)
        return self.make_get_request(
            "/alerts?{query_string}".format(query_string=urlencode(params)),
            end_point=self.DEFAULT_END_POINT)

    def active_alerts(self, count=False, **params):
        """Active alerts endpoints.

        Response in this method should not be modified.
        In this way, we can keep track of changes made by NOAA through
        functional tests @todo(paulokuong) later on.

        Args:
            count (bool): True to hit /alerts/active/count.
            zone_id (str): a valid zone, see list in counts endpoint.
            area (str): a valid area, see list in counts endpoint.
            region (str): a valid region, see list in counts endpoint
        Returns:
            json: json response from api.
        """

        if count:
            return self.make_get_request(
                "/alerts/count",
                end_point=self.DEFAULT_END_POINT)
        if 'zone_id' in params:
            return self.make_get_request(
                "/alerts/active/zone/{zoneId}".format(
                    zoneId=params['zone_id']),
                end_point=self.DEFAULT_END_POINT)
        if 'area' in params:
            return self.make_get_request(
                "/alerts/active/area/{area}".format(
                    area=params['area']),
                end_point=self.DEFAULT_END_POINT)
        if 'region' in params:
            return self.make_get_request(
                "/alerts/active/region/{region}".format(
                    region=params['region']),
                end_point=self.DEFAULT_END_POINT)
        return self.make_get_request(
            "/alerts/active",
            end_point=self.DEFAULT_END_POINT)
